"""Class for generateing a cross-section from a path of points.

Currently only 2 points can make up a path as in a cross-section.
"""
from typing import Tuple

import geopandas as gpd
import numpy as np
import numpy.typing as npt
import pandas as pd
from shapely.geometry import LineString
from shapely.geometry import Point


class PathGen:
    """Pathgen class."""

    def __init__(self: "PathGen", path_geom: gpd.GeoDataFrame, ny: int) -> None:
        """Init PathGen.

        Args:
            path_geom (gpd.GeoDataFrame): [description]
            ny (int): [description]
        """
        # print(path_geom, ny)
        self.path_geom = path_geom
        self.width = self.path_geom.geometry[0].length
        if ny % 2 == 0:
            ny += 1
        self.ny = ny
        self.x = np.zeros(self.ny, dtype=np.double)
        self.y = np.zeros(self.ny, dtype=np.double)
        self.int_path = None
        self.__buildpath()

    def __buildpath(self: "PathGen") -> None:
        line = self.path_geom.geometry[0]
        spacing = line.length / self.ny
        # print(line, spacing)
        d = 0.0
        index = 0
        while d < self.width and index < self.ny:
            point = line.interpolate(d)
            self.x[index] = point.x
            self.y[index] = point.y
            d += spacing
            index += 1

    def get_xs(self: "PathGen") -> gpd.GeoDataFrame:
        """Get resulting cross-section.

        Returns:
            Geopandas Dataframe: Return a Geopandas DataFrame of resulting cross-section
        """
        points = gpd.GeoSeries(map(Point, zip(self.x, self.y)))
        ls = LineString((points.to_list()))
        # print(ls)
        d = {0: {"name": "section-path", "geometry": ls}}
        df = pd.DataFrame.from_dict(d, orient="index")
        gdf = gpd.GeoDataFrame(df, geometry=df.geometry, crs=self.path_geom.crs)
        return gdf

    def get_xs_points(
        self: "PathGen",
    ) -> Tuple[npt.NDArray[np.double], npt.NDArray[np.double]]:
        """Get cross-sectin points.

        Returns:
            numpy arrays: Returns Numpy arrays of x and y points in cross-section.
        """
        return self.x, self.y



"""Pygeoapi instance of NLDI get cross-section at point."""
from __future__ import annotations

import logging
import time

from geopandas.geodataframe import GeoDataFrame
from pygeoapi.process.base import BaseProcessor

from nldi_xstool.nldi_xstool import getxsatpoint

# from typing import Any
# from typing import Tuple

LOGGER = logging.getLogger(__name__)

PROCESS_METADATA = {
    "version": "0.1.0",
    "id": "nldi-xsatpoint",
    "title": "NLDI xsatpoint process",
    "description": "NLDI xsatpoint process",
    "keywords": ["NLDI xsatpoint"],
    "links": [
        {
            "type": "text/html",
            "rel": "canonical",
            "title": "information",
            "href": "https://example.org/process",
            "hreflang": "en-US",
        }
    ],
    "inputs": [
        {
            "id": "lat",
            "title": "lat",
            "input": {
                "literalDataDomain": {
                    "dataType": "float",
                    "valueDefinition": {"anyValue": True},
                }
            },
            "minOccurs": 1,
            "maxOccurs": 1,
        },
        {
            "id": "lon",
            "title": "lon",
            "input": {
                "literalDataDomain": {
                    "dataType": "float",
                    "valueDefinition": {"anyValue": True},
                }
            },
            "minOccurs": 1,
            "maxOccurs": 1,
        },
        {
            "id": "width",
            "title": "width",
            "input": {
                "literalDataDomain": {
                    "dataType": "float",
                    "valueDefinition": {"anyValue": True},
                }
            },
            "minOccurs": 1,
            "maxOccurs": 1,
        },
        {
            "id": "numpts",
            "title": "numpts",
            "input": {
                "literalDataDomain": {
                    "dataType": "int",
                    "valueDefinition": {"anyValue": True},
                }
            },
            "minOccurs": 1,
            "maxOccurs": 1,
        },
    ],
    "outputs": [
        {
            "id": "nldi-xsatpoint-response",
            "title": "output nldi-xsatpoint",
            "output": {"formats": [{"mimeType": "application/json"}]},
        }
    ],
    "example": {
        "inputs": [
            {"id": "lat", "value": "40.2684", "type": "text/plain"},
            {"id": "lon", "value": "-103.80119", "type": "text/plain"},
            {"id": "width", "value": "1000.0", "type": "text/plain"},
            {"id": "numpts", "value": "101", "type": "text/plain"},
        ]
    },
}


class NLDIxsatpointProcessor(BaseProcessor):  # type: ignore
    """NLDI Get Cross-sectin at Point."""

    def __init__(self, provider_def):  # type: ignore
        """Initialize object.

        :param provider_def: provider definition
        :returns: pygeoapi.process.nldi_delineate.NLDIDelineateProcessor
        """
        BaseProcessor.__init__(self, provider_def, PROCESS_METADATA)

    def execute(self, data):  # type: ignore
        """Execute processor."""
        mimetype = "application/json"
        # reformat data into dict with just needed values
        newdata = {d["id"]: d["value"] for d in data}
        lat = float(newdata.get("lat"))  # type: ignore
        lon = float(newdata.get("lon"))  # type: ignore
        numpts = int(newdata.get("numpts"))  # type: ignore
        width = float(newdata.get("width"))  # type: ignore

        # print(lat, lon, width, numpts)

        timebefore = time.perf_counter()

        # print("before function")
        results = GeoDataFrame(getxsatpoint([lon, lat], numpts, width))
        # print("after function")
        # print(results)

        timeafter = time.perf_counter()
        totaltime = timeafter - timebefore
        print("Total Time:", totaltime)

        # outputs = [
        #     {"id": "nldi-xsatpoint-response", "value": results.__geo_interface__}
        # ]
        # print(results)
        return mimetype, results.__geo_interface__

    def __repr__(self):  # type: ignore
        """Get representation."""
        return "<NLDIxsatpointProcessor> {}".format("get xsatpoint")



"""Tools for extending USGS measured Bathymetry (preliminary)."""
from pathlib import Path

import geopandas as gpd
import numpy as np
import pandas as pd
from geopandas.array import points_from_xy
from shapely.geometry import Point

from .nldi_xstool import getxsatendpts


class ExtADCPBathy:
    """Class to facilitate extending USGS measured bathymetry cross-sections."""

    def __init__(
        self: "ExtADCPBathy",
        file: str,
        dist: float,
        lonstr: str,
        latstr: str,
        estr: str,
        acrs: str,
    ) -> None:
        """Init ExtADCPBathy class.

        Parameters
        ----------
        self : ExtADCPBathy
            [description]
        file : str
            [description]
        dist : float
            [description]
        lonstr : str
            [description]
        latstr : str
            [description]
        estr : str
            [description]
        acrs : str
            [description]
        """
        self.afile = Path(file)
        assert self.afile.exists()  # noqa S101
        self.adata = pd.read_csv(self.afile)
        self.agdf2 = gpd.GeoDataFrame(
            self.adata, geometry=points_from_xy(self.adata[lonstr], self.adata[latstr])
        )
        self.crs = acrs
        self.agdf = self.agdf2.set_crs(acrs)
        self.agdf["elevation"] = self.adata[estr]

        # convert to albers conic for geometry calculations
        self.agdf = self.agdf.to_crs("epsg:5071")

        # clean up dataframe by deleting old names
        del self.agdf[lonstr]
        del self.agdf[latstr]
        del self.agdf[estr]

        self.ext = dist
        self.xs_complete = gpd.GeoDataFrame()

        self._buildgeom()

    def _distance(self: "ExtADCPBathy", p1t: Point, p2t: Point) -> np.float64:
        """Convenience function to calc distance between 2 points."""
        return np.float64(np.sqrt(np.sum(np.square(np.array(p2t) - np.array(p1t)))))

    def _buildgeom(self: "ExtADCPBathy") -> None:
        npts = len(self.agdf)
        p1 = np.array(self.agdf.geometry[0])  # type: ignore
        p2 = np.array(self.agdf.geometry[npts - 1])  # type: ignore
        dx = p2[0] - p1[0]
        dy = p2[1] - p1[1]
        theta = np.arctan2(dy, dx)
        length = np.sqrt(np.sum(np.square(p2 - p1)))
        p2ny = p1[1] + (length + self.ext + 1) * np.sin(theta)
        p2nx = p1[0] + (length + self.ext + 1) * np.cos(theta)
        p1ny = p1[1] - (self.ext + 1) * np.sin(theta)
        p1nx = p1[0] - (self.ext + 1) * np.cos(theta)

        df_pre = pd.DataFrame(
            [{"lon": p1nx, "lat": p1ny}, {"lon": p1[0], "lat": p1[1]}]
        )

        df_post = pd.DataFrame(
            [{"lon": p2[0], "lat": p2[1]}, {"lon": p2nx, "lat": p2ny}]
        )

        gdf_pre = gpd.GeoDataFrame(
            df_pre, geometry=gpd.points_from_xy(df_pre.lon, df_pre.lat)
        ).set_crs("epsg:5071")
        # del gdf_pre['distance']
        gdf_pre_geo = gdf_pre.to_crs("epsg:4326")

        gdf_post = gpd.GeoDataFrame(
            df_pre, geometry=gpd.points_from_xy(df_post.lon, df_post.lat)
        ).set_crs("epsg:5071")
        # del gdf_post['distance']
        gdf_post_geo = gdf_post.to_crs("epsg:4326")

        new_df_pre = self._getxs(gdf_pre_geo, int(self.ext), 1).to_crs("epsg:5071")
        new_df_post = self._getxs(gdf_post_geo, int(self.ext), 1).to_crs("epsg:5071")
        del new_df_pre["distance"]
        del new_df_post["distance"]
        new_df_pre["code"] = "0"
        new_df_post["code"] = "0"
        self.agdf["code"] = "1"

        self.xs_complete = gpd.GeoDataFrame(
            pd.concat([new_df_pre, self.agdf, new_df_post], ignore_index=True),
            crs="epsg:5071",
        )

        self.xs_complete["station"] = 0.0
        for index, v in self.xs_complete.iterrows():
            if index == 0:
                p1 = v.geometry
            self.xs_complete.at[index, "station"] = self._distance(p1, v.geometry)

    def _getxs(
        self: "ExtADCPBathy", gdf: gpd.GeoDataFrame, numpts: int, res: int
    ) -> gpd.GeoDataFrame:
        """Convenience wrapper for nldi-xstool.getxsatendpts.

        Parameters
        ----------
        self : ExtADCPBathy
            [description]
        gdf : gpd.GeoDataFrame
            [description]
        numpts : int
            [description]
        res : int
            [description]

        Returns
        -------
        gpd.GeoDataFrame
            [description]
        """
        return getxsatendpts(
            [
                (gdf.geometry[0].x, gdf.geometry[0].y),
                (gdf.geometry[1].x, gdf.geometry[1].y),
            ],
            numpts=numpts,
            crs="epsg:4326",
            res=res,
        )

    def get_xs_complete(self: "ExtADCPBathy") -> gpd.GeoDataFrame:
        """Return extened cross-section.

        Parameters
        ----------
        self : ExtADCPBathy
            [description]

        Returns
        -------
        [type]
            [description]
        """
        return self.xs_complete.to_crs(self.crs)



"""Function to interpolate a tensioned spline based on user defined x,y points.

Code based on Fortran code developed by Jonathan Nelson (jmn@usgs.gov) at USGS.
"""
import numpy as np
import numpy.typing as npt


def tspline(
    x: npt.NDArray[np.double],
    y: npt.NDArray[np.double],
    n: int,
    xout: npt.NDArray[np.double],
    yout: npt.NDArray[np.double],
    iout: int,
    sigma: float,
    yp: npt.NDArray[np.double],
    temp: npt.NDArray[np.double],
) -> None:
    """Tension spline generator.

    Parameters
    ----------
    x : npt.NDArray[np.double]
        [description]
    y : npt.NDArray[np.double]
        [description]
    n : int
        [description]
    xout : npt.NDArray[np.double]
        [description]
    yout : npt.NDArray[np.double]
        [description]
    iout : int
        [description]
    sigma : float
        [description]
    yp : npt.NDArray[np.double]
        [description]
    temp : npt.NDArray[np.double]
        [description]
    """
    nm1 = n - 2
    np1 = n
    delx1 = x[1] - x[0]
    dx1 = (y[1] - y[0]) / delx1
    delx2 = x[2] - x[1]
    delx12 = x[2] - x[0]
    c1 = -(delx12 + delx1) / delx12 / delx1
    c2 = delx12 / delx1 / delx2
    c3 = -delx1 / delx12 / delx2
    slpp1 = c1 * y[0] + c2 * y[1] + c3 * y[2]
    deln = x[n - 1] - x[nm1]
    delnm1 = x[nm1] - x[n - 3]
    delnn = x[n - 1] - x[n - 3]
    c1 = (delnn + deln) / delnn / deln
    c2 = -delnn / deln / delnm1
    c3 = deln / delnn / delnm1
    slppn = c3 * y[n - 3] + c2 * y[nm1] + c1 * y[n - 1]
    sigmap = np.fabs(sigma) * (float(n - 2)) / (x[n - 1] - x[0])
    dels = sigmap * delx1
    exps = np.exp(dels)
    sinhs = 0.5 * (exps - 1.0 / exps)
    sinhin = 1.0 / (delx1 * sinhs)
    diag1 = sinhin * (dels * 0.5 * (exps + 1.0 / exps) - sinhs)
    diagin = 1.0 / diag1
    yp[0] = diagin * (dx1 - slpp1)
    spdiag = sinhin * (sinhs - dels)
    temp[0] = diagin * spdiag
    dx2 = 0
    for i in range(0, nm1):
        delx2 = x[i + 1] - x[i]
        dx2 = (y[i + 1] - y[i]) / delx2
        dels = sigmap * delx2
        exps = np.exp(dels)
        sinhs = 0.5 * (exps - 1.0 / exps)
        sinhin = 1.0 / (delx2 * sinhs)
        diag2 = sinhin * (dels * (0.5 * (exps + 1.0 / exps)) - sinhs)
        diagin = 1.0 / (diag1 + diag2 - spdiag * temp[i - 1])
        yp[i] = diagin * (dx2 - dx1 - spdiag * yp[i - 1])
        spdiag = sinhin * (sinhs - dels)
        temp[i] = diagin * spdiag
        dx1 = dx2
        diag1 = diag2
    diagin = 1.0 / (diag1 - spdiag * temp[nm1])
    yp[n - 1] = diagin * (slppn - dx2 - spdiag * yp[nm1])
    for i in range(1, n):
        ibak = np1 - i - 1
        yp[ibak] = yp[ibak] - temp[ibak] * yp[ibak + 1]
    a = x[0]
    b = x[1]
    nj = 1
    for i in range(0, iout):
        if xout[i] > b:
            while xout[i] > b:
                a = b
                nj = nj + 1
                b = x[nj]

        del1 = xout[i] - a
        del2 = b - xout[i]
        dels = b - a
        exps1 = np.exp(sigmap * del1)
        sinhd1 = 0.5 * (exps1 - 1.0 / exps1)
        exps = np.exp(sigmap * del2)
        sinhd2 = 0.5 * (exps - 1.0 / exps)
        exps = exps * exps1
        sinhs = 0.5 * (exps - 1.0 / exps)
        yout[i] = (yp[nj] * sinhd1 + yp[nj - 1] * sinhd2) / sinhs + (
            (y[nj] - yp[nj]) * del1 + (y[nj - 1] - yp[nj - 1]) * del2
        ) / dels



"""Main module."""
# from nldi_xstool.cli import xsatendpts
import sys
from typing import Any
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

import geopandas as gpd
import numpy as np
import numpy.typing as npt
import pandas as pd
import py3dep
import requests
from pynhd import NLDI
from shapely.geometry import LineString
from shapely.geometry import Point

from .PathGen import PathGen
from .XSGen import XSGen


def dataframe_to_geodataframe(
    df: pd.DataFrame, crs: str
) -> gpd.GeoDataFrame:  # noqa D103
    """Convert pandas Dataframe to Geodataframe."""
    geometry = [Point(xy) for xy in zip(df.x, df.y)]
    df = df.drop(["x", "y"], axis=1)
    gdf = gpd.GeoDataFrame(df, geometry=geometry, crs=crs)
    return gdf


def get_elev_along_path(
    path: List[Tuple[float, float]],
    numpts: int,
    crs: Optional[str] = "epsg:4326",
    file: Optional[str] = None,
    res: Optional[int] = 10,
) -> Union[int, gpd.GeoDataFrame]:
    """Get elevation at user-defined path.

    Note 1:
        Cross-section is interpolated using 3DEP elevation services which does not
        include bathymetric data.

    Note 2:
        The parameter res specifices the spatial resolution of the 3DEP data that is
        interpolated to return elevation values of the cross-section points.  It does
        not specify the native spatial resolution of the 3DEP data.  If using this as a
        package the query_dems function in nldi-xstool.ancillary can be used to discover
        the available native resolution of the 3DEP data at the bounding box of the
        cross-section.

    Args:
        path (List[Tuple[float, float]]): List of tuples containing coordinate pairs
            of the path, for example: [(x1,y1), (x2,y2), (x3, y3)]
        numpts (int): Number of points to interpolate along path.
        crs (Optional[str], optional): crs of input data. Defaults to "epsg:4326".
        file (Optional[str], optional): path/to/file.json. Defaults to None and returns
            a GeoDataFrame.
        res (Optional[int], optional): Spatial resolution of 3DEP Dem data. Defaults to
            10 which is available throughout CONUS.

    Returns:
        Union[int, gpd.GeoDataFrame]: If a file is specified it writes a json file and
            returns 0, otherwise it returns the cross-section in a GeoDataFrame.
    """
    lnst = []
    for pt in path:
        lnst.append(Point(pt[0], pt[1]))

    # print(ls1)
    d = {"name": ["xspath"], "geometry": [LineString(lnst)]}
    gpd_pth = gpd.GeoDataFrame(d, crs=crs)
    # print(gpd_pth)
    gpd_pth.to_crs(epsg=3857, inplace=True)
    # print(gpd_pth)
    xs = PathGen(path_geom=gpd_pth, ny=numpts)
    xs_line = xs.get_xs()
    # print(xs_line.head())
    # print(xs_line.total_bounds, xs_line.bounds)
    bb = xs_line.total_bounds - ((100.0, 100.0, -100.0, -100.0))
    # print('before dem', bb)
    dem = py3dep.get_map(
        "DEM", tuple(bb), resolution=res, geo_crs="EPSG:3857", crs="epsg:3857"
    )
    # print('after dem')
    x, y = xs.get_xs_points()
    dsi = dem.interp(x=("z", x), y=("z", y))
    pdsi = dsi.to_dataframe()
    gpdsi = dataframe_to_geodataframe(pdsi, crs="epsg:3857")
    gpdsi["distance"] = _get_dist_path(gpdsi)
    gpdsi.to_crs(epsg=4326, inplace=True)
    if file:
        with open(file, "w") as f:
            f.write(gpdsi.to_json())
            f.close()
            return 0
    else:
        return gpdsi


def getxsatendpts(
    path: List[Tuple[float, float]],
    numpts: int,
    crs: Optional[str] = "epsg:4326",
    file: Optional[str] = None,
    res: Optional[int] = 10,
) -> Union[int, gpd.GeoDataFrame]:
    """Get cross-section at user defined endpoints.

    Note 1:
        Cross-section is interpolated using 3DEP elevation services which does not
        include bathymetric data.

    Note 2:
        The parameter res specifices the spatial resolution of the 3DEP data that is
        interpolated to return elevation values of the cross-section points.  It does
        not specify the native spatial resolution of the 3DEP data.  If using this as a
        package the query_dems function in nldi-xstool.ancillary can be used to discover
        the available native resolution of the 3DEP data at the bounding box of the
        cross-section.

    Args:
        path (List[Tuple[float, float]]): List of tuples containing coordinate pairs
            of the end-points in order from river-left to river-right,
            for example: [(x1,y1), (x2,y2)]
        numpts (int): Number of points to interpolate along path.
        crs (Optional[str], optional): crs of input data. Defaults to "epsg:4326".
        file (Optional[str], optional): path/to/file.json. Defaults to None and returns
            a GeoDataFrame.
        res (Optional[int], optional): Spatial resolution of 3DEP Dem data. Defaults to
            10 which is available throughout CONUS.

    Returns:
        Union[int, gpd.GeoDataFrame]: If a file is specified it writes a json file and
            returns 0, otherwise it returns the cross-section in a GeoDataFrame.
    """
    lnst = []
    for pt in path:
        lnst.append(Point(pt[0], pt[1]))

    # print(ls1)
    d = {"name": ["xspath"], "geometry": [LineString(lnst)]}
    gpd_pth = gpd.GeoDataFrame(d, crs=crs)
    # print(gpd_pth)
    # gpd_pth.set_crs(epsg=4326, inplace=True)
    gpd_pth.to_crs(epsg=3857, inplace=True)
    # print(gpd_pth)
    xs = PathGen(path_geom=gpd_pth, ny=numpts)
    xs_line = xs.get_xs()
    # print(xs_line.head())
    # print(xs_line.total_bounds, xs_line.bounds)
    bb = xs_line.total_bounds - ((100.0, 100.0, -100.0, -100.0))
    # print('before dem', bb)
    dem = py3dep.get_map(
        "DEM", tuple(bb), resolution=res, geo_crs="EPSG:3857", crs="epsg:3857"
    )

    # print('after dem')
    x, y = xs.get_xs_points()
    dsi = dem.interp(x=("z", x), y=("z", y))
    pdsi = dsi.to_dataframe()

    gpdsi = dataframe_to_geodataframe(pdsi, crs="epsg:3857")
    gpdsi["distance"] = _get_dist(gpdsi)
    gpdsi.to_crs(epsg=4326, inplace=True)
    if file:
        with open(file, "w") as f:
            f.write(gpdsi.to_json())
            f.close()
            return 0
    else:
        return gpdsi


def _get_dist(gdf: gpd.GeoDataFrame) -> Any:
    data = gdf.to_crs(epsg=5071)
    x1 = data["geometry"].x.values[0] - data["geometry"].x.values[:]
    y1 = data["geometry"].y.values[0] - data["geometry"].y.values[:]
    return np.hypot(x1, y1)


def _get_dist_path(gdf: gpd.GeoDataFrame) -> npt.NDArray[np.float64]:
    data = gdf.to_crs(epsg=5071)
    distance = []
    for index, g in enumerate(data.geometry):
        if index == 0:
            p1 = g
            p2 = g
            distance.append(0.0)
        else:
            p1 = p2
            p2 = g
            x1 = p2.x - p1.x
            y1 = p2.y - p1.y
            sumd = distance[-1]
            distance.append(sumd + np.hypot(x1, y1))
    return np.array(distance)


def getxsatpoint(
    point: List[float],
    numpoints: int,
    width: float,
    file: Optional[str] = None,
    res: Optional[int] = 10,
) -> Optional[Union[gpd.GeoDataFrame, None]]:
    """Get cross-section at nearest stream-segment and closest intersecton to point.

    Function uses the U.S. Geological Survey's NLDI to find the nearest NHD stream-
    segment, and generate a cross-section perpendicular to the nearest intersection
    of the point and stream-segment.

    Note 1:
        Cross-section is interpolated using 3DEP elevation services which does not
        include bathymetric data.

    Note 2:
        The parameter res specifices the spatial resolution of the 3DEP data that is
        interpolated to return elevation values of the cross-section points.  It does
        not specify the native spatial resolution of the 3DEP data.  If using this as a
        package the query_dems function in nldi-xstool.ancillary can be used to discover
        the available native resolution of the 3DEP data at the bounding box of the
        cross-section.

    Args:
        point (List[float]): Point of interest, for example [x,y]
        numpoints (int): Number of points to interpolate along path.
        width (float): Width of the cross-section centered on the stream-segment
        file (Optional[str], optional): path/to/file.json. Defaults to None and returns
            a GeoDataFrame.
        res (Optional[int], optional): Spatial resolution of 3DEP Dem data. Defaults to
            10 which is available throughout CONUS.

    Returns:
        Union[int, gpd.GeoDataFrame]: If a file is specified it writes a json file and
            returns 0, otherwise it returns the cross-section in a GeoDataFrame.
    """
    # print(file, type(file))
    # tpoint = f'POINT({point[1]} {point[0]})'
    df = pd.DataFrame(
        {"pointofinterest": ["this"], "Lat": [point[1]], "Lon": [point[0]]}
    )
    gpd_pt = gpd.GeoDataFrame(df, geometry=gpd.points_from_xy(df.Lon, df.Lat))
    gpd_pt.set_crs(epsg=4326, inplace=True)
    gpd_pt.to_crs(epsg=3857, inplace=True)
    try:
        comid = __get_cid_from_lonlat(point)
    except Exception as ex:  # pragma: no cover
        # print(f'Error: {ex} unable to find comid - check lon lat coords')
        sys.exit(f"Error: {ex} unable to find comid - check lon lat coords")
    # print(f'comid = {comid}')
    strm_seg = NLDI().getfeature_byid("comid", comid).to_crs("epsg:3857")
    xs = XSGen(point=gpd_pt, cl_geom=strm_seg, ny=numpoints, width=width, tension=10.0)
    xs_line = xs.get_xs()
    # print(comid, xs_line)
    # get topo polygon with buffer to ensure there is enough topography to interpolate
    # xs line with coarsest DEM (30m) 100. m should
    bb = xs_line.total_bounds - ((100.0, 100.0, -100.0, -100.0))
    dem = py3dep.get_map(
        "DEM", tuple(bb), resolution=res, geo_crs="EPSG:3857", crs="epsg:3857"
    )
    x, y = xs.get_xs_points()
    dsi = dem.interp(x=("z", x), y=("z", y))
    pdsi = dsi.to_dataframe()
    gpdsi = dataframe_to_geodataframe(pdsi, crs="epsg:3857")
    gpdsi["distance"] = _get_dist(gpdsi)
    gpdsi.to_crs(epsg=4326, inplace=True)
    if file:
        with open(str(file), "w") as f:
            f.write(gpdsi.to_json())
            f.close()
        return None
    else:
        return gpdsi  # pragma: no cover


def __lonlat_to_point(lon: float, lat: float) -> Point:  # noqa D103
    return Point(lon, lat)


def __get_cid_from_lonlat(point: List[float]) -> str:
    # print(point)
    pt = __lonlat_to_point(point[0], point[1])

    location = pt.wkt
    location = f"POINT({point[0]} {point[1]})"
    domain = "https://labs.waterdata.usgs.gov"
    path = "/api/nldi/linked-data/comid/position?f=json&coords="
    base_url = domain + path
    url = base_url + location
    # print(url)
    try:
        response = requests.get(url)
        # print(f'this is the response: {response}')
        response.raise_for_status()
        jres = response.json()
        comid = jres["features"][0]["properties"]["comid"]

    except requests.exceptions.RequestException as err:  # pragma: no cover
        print("OOps: Something Else", err)
    except requests.exceptions.HTTPError as errh:  # pragma: no cover
        print("Http Error:", errh)
    except requests.exceptions.ConnectionError as errc:  # pragma: no cover
        print("Error Connecting:", errc)
    except requests.exceptions.Timeout as errt:  # pragma: no cover
        print("Timeout Error:", errt)
    except Exception as ex:  # pragma: no cover
        raise ex

    return str(comid)



"""Created on Tue May  5 16:26:25 2015.

@author: mweier
"""
# -*- coding: utf-8 -*-
import numpy as np
from numba import jit

# noqa N803


@jit
def channelbuilder(wsdepth, rightss, leftss, widthbottom):  # type: ignore
    """Channel builder.

    Builds trapziodal channel station/elevation array given depth,
    right side slope, left side slope, and bottom width.
    """
    lefttoe = wsdepth * 1.25 * leftss
    righttoe = wsdepth * 1.25 * rightss
    staelev = np.array(
        [
            (0.0, wsdepth * 1.25),
            (lefttoe, 0.0),
            (lefttoe + widthbottom, 0.0),
            (lefttoe + widthbottom + righttoe, wsdepth * 1.25),
        ]
    )  # type: ignore
    return staelev


def lineintersection(line1, line2):  # type: ignore
    """lineintersection.

    Args:
        line1 ([type]): [description]
        line2 ([type]): [description]

    Returns:
        [type]: [description]
    """
    xdiff = (line1[0][0] - line1[1][0], line2[0][0] - line2[1][0])
    ydiff = (line1[0][1] - line1[1][1], line2[0][1] - line2[1][1])

    def det(a, b):  # type: ignore
        return a[0] * b[1] - a[1] * b[0]

    div = det(xdiff, ydiff)  # type: ignore
    if div == 0:
        x = y = np.nan
        #        print 'lines do not intersect'
        return x, y

    d = (det(*line1), det(*line2))  # type: ignore
    x = det(d, xdiff) / div  # type: ignore
    y = det(d, ydiff) / div  # type: ignore
    return x, y


@jit
def polygon_area(corners):  # type: ignore
    """polygon_area.

    Args:
        corners ([type]): [description]

    Returns:
        [type]: [description]
    """
    area = 0.0
    for i in range(len(corners)):
        j = (i + 1) % len(corners)
        area += corners[i][0] * corners[j][1]
        area -= corners[j][0] * corners[i][1]
    area = abs(area) / 2.0
    return area


@jit
def channel_perimeter(corners):  # type: ignore
    """Calculate channel perimeter.

    Parameters
    ----------
    corners : [type]
        [description]

    Returns
    -------
    [type]
        [description]
    """
    p = 0.0
    for i in range(len(corners) - 1):
        p += np.sqrt(
            (
                np.power((corners[i + 1][0] - corners[i][0]), 2)
                + np.power((corners[i + 1][1] - corners[i][1]), 2)
            )
        )
    return p


def flow_est(wselev, n, slope, staelev, units):  # type: ignore
    """[summary].

    Estimates uniform flow using the Manning equation for
    a user defined trapziodal channel or a manually defined channel using
    a station/elevation file

    Args:
        wselev ([type]): [description]
        n ([type]): [description]
        slope ([type]): [description]
        staelev ([type]): [description]
        units ([type]): [description]

    Returns:
        [type]: [description]
    """
    if units == "m":
        const = 1.0
    else:
        const = 1.49

    intersectlist = []
    for i in range(0, len(staelev)):
        x, y = lineintersection(
            (staelev[i - 1], staelev[i]),
            ([staelev[0][0], wselev], [staelev[-1][0], wselev]),
        )  # type: ignore
        #         print(x,y)
        if x >= staelev[i - 1][0] and x <= staelev[i][0] and abs(y - wselev) < 0.01:
            #             print (x,y)
            intersectlist.append((x, y))
        else:
            #             print ('line segments do not intersect')
            pass

    try:
        intersectarray = np.array(intersectlist)  # type: ignore
        intersectarray = intersectarray[intersectarray[:, 0].argsort()]
        # print 'more than two points intersect'
        staminelev = staelev[np.where(staelev[:, 1] == min(staelev[:, 1]))][0][0]
        startpoint = intersectarray[np.where(intersectarray[:, 0] < staminelev)][-1]
        endpoint = intersectarray[np.where(intersectarray[:, 0] > staminelev)][0]
        intersectarray = np.vstack([startpoint, endpoint])
    except Exception as e:
        print(e)
        return 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0

    stamin = np.min(intersectarray[:, 0])
    stamax = np.max(intersectarray[:, 0])

    thalweig = staelev[np.where(staelev[:, 1] == np.min(staelev[:, 1]))]

    minelev = thalweig[:, 1][0]
    maxdepth = wselev - minelev

    if len(intersectarray) < 2:
        return 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0

    staelevtrim = np.vstack([intersectarray[0], staelev, intersectarray[1]])  # type: ignore
    # staelevtrim = staelevtrim[staelevtrim[:,0].argsort()]
    staelevtrim = staelevtrim[
        np.where((staelevtrim[:, 0] >= stamin) & (staelevtrim[:, 0] <= stamax))
    ]

    area = polygon_area(staelevtrim)
    r = area / channel_perimeter(staelevtrim)
    v = (const / n) * np.power(r, (2.0 / 3.0)) * np.sqrt(slope)
    q = v * area
    topwidth = stamax - stamin
    xground = staelev[:, 0]
    yground = staelev[:, 1]
    yground0 = np.ones(len(xground)) * np.min(yground)
    xwater = staelevtrim[:, 0]
    ywater = np.ones(len(xwater)) * wselev
    ywater0 = staelevtrim[:, 1]
    args = (
        r,
        area,
        topwidth,
        q,
        v,
        maxdepth,
        xground,
        yground,
        yground0,
        xwater,
        ywater,
        ywater0,
    )
    return args



"""Command-line interface."""
import sys
from typing import Any
from typing import Tuple

import click
import shapely

from .nldi_xstool import getxsatendpts
from .nldi_xstool import getxsatpoint

# from typing import List

# 3dep resolution
resdict = {"1m": 1, "3m": 3, "5m": 5, "10m": 10, "30m": 30, "60m": 60}

# from:
# https://gis.stackexchange.com/questions/363221/
# how-do-you-validate-longitude-and-latitude-coordinates-in-python/378885#378885


def valid_lonlat(ctx, param, value):  # type: ignore
    """This validates a lat and lon.

    This validates a lat and lonpoint can be located
    in the bounds of the WGS84 CRS, after wrapping the
    longitude value within [-180, 180).

    :param lon: a longitude value
    :param lat: a latitude value
    :return: (lon, lat) if valid, None otherwise
    """
    print(f"latlon: {param, value}")

    lon = float(value[0])
    lat = float(value[1])
    # Put the longitude in the range of [0,360):
    lon %= 360
    # Put the longitude in the range of [-180,180):
    if lon >= 180:
        lon -= 360
    lon_lat_point = shapely.geometry.Point(lon, lat)
    lon_lat_bounds = shapely.geometry.Polygon.from_bounds(
        xmin=-180.0, ymin=-90.0, xmax=180.0, ymax=90.0
    )
    # return lon_lat_bounds.intersects(lon_lat_point)
    # would not provide any corrected values
    try:
        if lon_lat_bounds.intersects(lon_lat_point):
            return tuple((lon, lat))
            # return lon, lat
    except ValueError:
        msg = f"Not valid lon lat pair: {lon, lat}"
        print(msg)
        # raise click.BadParameter(msg)


class NLDIXSTool:
    """Simple class to handle global crs."""

    def __init__(self: "NLDIXSTool") -> None:
        """Init NLDIXSTool."""
        self.out_crs = "epsg:4326"

    def setoutcrs(self: "NLDIXSTool", out_crs: str = "epsg:4326") -> None:
        """Set the CRS for output.

        Args:
            out_crs (str): Set the CRS string for projection of output. Defaults to "epsg:4326".
        """
        self.out_crs = out_crs

    def outcrs(self: "NLDIXSTool") -> str:
        """Get the output CRS.

        Returns:
            str: The epsg crs code.
        """
        return self.out_crs

    def __repr__(self: "NLDIXSTool") -> str:
        """Representation.

        Returns:
            str: Representation ouput.
        """
        return f"NLDI_XSTool {self.out_crs}"


pass_nldi_xstool = click.make_pass_decorator(NLDIXSTool)


@click.group()
@click.option(
    "--outcrs",
    default="epsg:4326",
    help="Projection CRS to return cross-section geometry: default is epsg:4326",
)  # type: ignore
@click.version_option("0.1")  # type: ignore
@click.pass_context
def main(ctx: Any, outcrs: str) -> int:
    """nldi-xstooln is a command line tool to for elevation-based services to the NLDI."""
    ctx.obj = NLDIXSTool()
    ctx.obj.setoutcrs(outcrs)
    return 0


# XS command at point with NHD


@main.command()
@click.option(
    "-f",
    "--file",
    default=None,
    type=str,
    help="enter path and filename for json ouput: path/to/file.json",
)  # type: ignore
@click.option(
    "-ll",
    "--lonlat",
    required=True,
    # type=valid_lonlat(float, float),
    type=tuple((float, float)),
    callback=valid_lonlat,  # type: ignore
    help="format lon,lat (x,y) as floats for example: -103.8011 40.2684",
)
@click.option(
    "-n", "--numpoints", default=101, type=int, help="number of points in cross-section"
)  # type: ignore
@click.option(
    "-w", "--width", default=1000.0, type=float, help="width of cross-section"
)  # type: ignore
@click.option(
    "-r",
    "--resolution",
    type=click.Choice(["1m", "3m", "5m", "10m", "30m", "60m"], case_sensitive=False),
    default="10m",
    help=(
        "Resolution of DEM used.  "
        "Note: 3DEP provides server side interpolatin given best available data"
    ),
)  # type: ignore
@click.option(
    "-v", "--verbose", default=False, type=bool, help="verbose ouput"
)  # type: ignore
@pass_nldi_xstool  # type: ignore
def xsatpoint(
    nldi_xstool: "NLDIXSTool",
    lonlat: Tuple[float, float],
    numpoints: int,
    width: float,
    resolution: str,
    file: str,
    verbose: bool,
) -> int:
    """Topographic cross-section at user-defined point.

    This function relies on the U.S. Geological Survey's NLDI and 3DEP program elevation
    services to return a topographic cross-section, at nearest NHD stream-segment
    based on user-defined point, width, number of points, and 3DEP spatial resolution.

    NOTE: 3DEP does not specifically include bathymetry.
    """
    coord = []
    coord.append(lonlat[0])
    coord.append(lonlat[1])
    x = lonlat[0]  # noqa DAR003
    y = lonlat[1]
    nl = "\n"
    if verbose:
        print(
            f"input={lonlat}, lat={x}, lon={y}, {nl} \
                npts={numpoints}, width={width}, resolution={resolution}, {nl} \
                crs={nldi_xstool.outcrs()}, {nl} \
                file={file}, {nl} \
                out_epsg={nldi_xstool.outcrs()}"
        )
    # print(tuple(latlon))
    xs = getxsatpoint(
        point=coord,
        numpoints=numpoints,
        width=width,
        file=file,
        res=resdict.get(resolution),
    )

    if not file:
        print(xs.to_json())  # type: ignore
    return 0


# XS command at user defined endpoints


@main.command()
@click.option(
    "-f",
    "--file",
    default=None,
    type=str,
    help="enter path and filename for json ouput: path/to/file.json",
)  # type: ignore
@click.option(
    "-s",
    "--startpt",
    required=True,
    type=tuple((float, float)),
    help="format x y pair as floats for example: -103.801134 40.267335",
)  # type: ignore
@click.option(
    "-e",
    "--endpt",
    required=True,
    type=tuple((float, float)),
    help="format x y pair as floats for example: -103.800787 40.272798 ",
)  # type: ignore
@click.option(
    "-c",
    "--crs",
    required=True,
    type=str,
    help="spatial reference of input data",
    default="epsg:4326",
)  # type: ignore
@click.option(
    "-n", "--numpoints", default=100, type=int, help="number of points in cross-section"
)  # type: ignore
@click.option(
    "-r",
    "--resolution",
    type=click.Choice(["1m", "3m", "5m", "10m", "30m", "60m"], case_sensitive=False),
    default="10m",
    help=(
        "Resolution of DEM used."
        + "  Note: 3DEP provides server side interpolatin given best available data"
    ),
)  # type: ignore
@click.option(
    "-v", "--verbose", default=False, type=bool, help="verbose ouput"
)  # type: ignore
@pass_nldi_xstool  # type: ignore
def xsatendpts(
    nldi_xstool: "NLDIXSTool",
    startpt: Tuple[float, float],
    endpt: Tuple[float, float],
    crs: str,
    numpoints: int,
    resolution: str,
    file: str,
    verbose: bool,
) -> int:
    """Topographic cross-section at user-defined end points.

    This function relies on the U.S. Geological Survey's 3DEP program elevation services
    to return a topographic cross-section based on user-defined end-points, number of
    points, and 3DEP spatial resolution.

    NOTE: 3DEP does not specifically include bathymetry.
    """
    x1 = startpt[0]
    y1 = startpt[1]
    x2 = endpt[0]
    y2 = endpt[1]
    nl = "\n"
    if verbose:
        print(
            f"input:  {nl}, \
            start: {startpt}, {nl}, \
            end: {endpt}, {nl}, \
            x1:{x1}, y1:{y1}, {nl}, \
            x2:{x2}, y2:{y2}, {nl}, \
            npts={numpoints}, {nl}, \
            resolution={resolution}, \
            input_crs={crs}, {nl}, \
            output_crs={nldi_xstool.outcrs()}  {nl}, \
            file={file}, {nl}, \
            verbose: {verbose} "
        )
    path = []
    path.append(startpt)
    path.append(endpt)
    # print(type(path))
    xs = getxsatendpts(
        path=path, numpts=numpoints, res=resdict.get(resolution), crs=crs, file=file
    )
    if not file:
        print(xs.to_json())  # type: ignore
    return 0


if __name__ == "__main__":
    sys.exit(main(prog_name="nldi-xstool"))  # pragma: no cover



"""Ancilarry."""
from typing import Any
from typing import List
from typing import Tuple

import geopandas as gpd
import requests
from dataretrieval import nwis
from shapely.geometry import LineString
from shapely.geometry import Point
from shapely.geometry import Polygon

from .ExtADCPBathy import ExtADCPBathy


def get_ext_bathy_xs(
    file: str, dist: float, lonstr: str, latstr: str, estr: str, acrs: str
) -> gpd.GeoDataFrame:
    """Extend a measured bathymetric cross-section to above water topography.

    Parameters
    ----------
    file : str
        Bathymetric survey file
    dist : float
        Distance to extend survey on both ends.
    lonstr : str
        String heading on longitude in file.
    latstr : str
        String heading on latitude in file.
    estr : str
        String heading on elevation in file.
    acrs : str
        String heading on crs code of measured lon/lat values, for example: espg:4326.

    Returns
    -------
    gpd.GeoDataFrame
        Geopandas dataframe of complete cross-section.
    """
    exs = ExtADCPBathy(
        file=file, dist=dist, lonstr=lonstr, latstr=latstr, estr=estr, acrs=acrs
    )

    return exs.get_xs_complete()


# The following function converts NGVD29 to NAVD88 if gage is in NGVD29 using
# NOAA NGS Vertcon service api
# https://www.ngs.noaa.gov/web_services/ncat/lat-long-height-service.shtml


def get_gage_datum(gagenum: str, verbose: bool = False) -> float:
    """Returns the datum of USGS gage in meters.

    Function uses the NOAA NGS Vertcon service to convert NGVD29 to NAVD88 if necessary.

    Parameters
    ----------
    gagenum : str
        USGS Gage number
    verbose : bool, optional
        Verbose output, by default False

    Returns
    -------
    float
        USGS Gage datum in meters
    """
    si = nwis.get_record(sites=gagenum, service="site")
    if si["alt_datum_cd"].values[0] == "NGVD29":
        # print('conversion')
        url = "https://www.ngs.noaa.gov/api/ncat/llh"
        lat_str = "dec_lat_va"
        lon_str = "dec_long_va"

        alt_str = "alt_va"
        indatum_str = "coord_datum_cd"
        outdatum_str = "NAD83(2011)"
        in_vert_dataum_str = "NGVD29"
        out_vert_dataum_str = "NAVD88"
        if f"{si[indatum_str].values[0]}" == "NAD83":
            indatum = "NAD83(2011)"
        else:
            indatum = f"{si[indatum_str].values[0]}"

        ohgt = float(si[alt_str].values[0]) * 0.3048

        tmplonstr = si[lon_str].values[0]

        if len(str(tmplonstr)) == 6:
            tstr = "0"
            tmplonstr = tstr + str(tmplonstr)

        payload = {
            "lat": f"{si[lat_str].values[0]}",
            "lon": f"{tmplonstr}",
            "orthoHt": repr(ohgt),
            "inDatum": indatum,
            "outDatum": outdatum_str,
            "inVertDatum": in_vert_dataum_str,
            "outVertDatum": out_vert_dataum_str,
        }
        r = requests.get(url, params=payload)
        resp = r.json()
        if verbose:
            print(f"{si[indatum_str].values[0]}")
            print(f"payload: {payload}")
            print(resp)
        return float(resp["destOrthoht"])
    else:
        # print('non-conversion')
        return float(si["alt_va"].values[0] * 0.3048)


# Resolution types and their respective IDs for the Rest Service
res_types = {
    "res_1m": 18,
    "res_3m": 19,
    "res_5m": 20,
    "res_10m": 21,
    "res_30m": 22,
    "res_60m": 23,
}
# res_types = {'res_1m': 1, 'res_3m': 2, 'res_5m': 3, 'res_10m': 4, 'res_30m': 5, 'res_60m': 6}
dim_order = {"latlon": 0, "lonlat": 1}
# Create a bounding box from any geo type
# 'width' is in meters. It is the width of the buffer to place around the input geometry


def make_bbox(shape_type: str, coords: List[Tuple[float, float]], width: int) -> Any:
    """Return a bounding box of the coordinates buffered by width.

    Parameters
    ----------
    shape_type : str
        Shapetype one of: point, line, polygon
    coords : List[Tuple]
        List of coordinate tuples
    width : int
        Width in meters

    Returns
    -------
    Any
        Bounding box of geometry (minx, miny, maxx, maxy)
    """
    if shape_type == "point":
        shape = Point(coords)

    elif shape_type == "line":
        shape = LineString(coords)

    elif shape_type == "polygon":
        shape = Polygon(coords)

    converted_width = convert_width(width)
    buff_shape = shape.buffer(converted_width)
    return buff_shape.bounds


# Convert the width of the buffer from meters to 'degree'
# This is NOT a precise conversion, just a quick overestimation
# Maybe fix this later


def convert_width(width: int) -> int:
    """Buffer bbox geometry with specified width.

    This is an esitmation and is not precise.

    Parameters
    ----------
    width : int
        Width in meters to bffer geometries bounding box.

    Returns
    -------
    int
        buffer in degrees
    """
    factor = 1 / 70000
    dist = factor * width
    return int(dist)


# Get request from arcgis Rest Services


def get_dem(bbox: Tuple[float, float, float, float], res_type: str) -> bool:
    """Esri map service query of 3DEPElevationIncex using bounding box.

    Return the True if res_type (Resolution of 3DEP elevation) intersects bounding box (bbox)

    Parameters
    ----------
    bbox : Tuple[float, float, float, float]
        (minx, miny, maxx, maxy)
    res_type : str
        Key in: res_types = {"res_1m": 18, "res_3m": 19,"res_5m": 20,
            "res_10m": 21,"res_30m": 22,"res_60m": 23,}

    Returns
    -------
    bool
        True if bbox intersects 3DEP elevation with res_type
    """
    minx = str(bbox[0])
    miny = str(bbox[1])
    maxx = str(bbox[2])
    maxy = str(bbox[3])
    res_id = res_types[res_type]
    return_val = False
    domain = "https://index.nationalmap.gov/"
    path = f"arcgis/rest/services/3DEPElevationIndex/MapServer/{res_id}/query"
    url = domain + path
    payload = {
        # "where": "",
        # "text": "",
        # "objectIds": "",
        # "time": "",
        "geometry": '{xmin:"'
        + minx
        + '",ymin:"'
        + miny
        + '",xmax:"'
        + maxx
        + '",ymax:"'
        + maxy
        + '",spatialReference:{wkid:4326}}',
        "geometryType": "esriGeometryEnvelope",
        "inSR": "EPSG:4326",
        "spatialRel": "esriSpatialRelIntersects",
        # "relationParam": "",
        # "outFields": "",
        "returnGeometry": "true",
        # "returnTrueCurves": "false",
        "maxAllowableOffset": "100",
        "geometryPrecision": "3",
        "outSR": "EPSG:4326",
        # "having": "",
        # "returnIdsOnly": "false",
        # "returnCountOnly": "false",
        # "orderByFields": "",
        # "groupByFieldsForStatistics": "",
        # "outStatistics": "",
        # "returnZ": "false",
        # "returnM": "false",
        # "gdbVersion": "",
        # "historicMoment": "",
        # "returnDistinctValues": "false",
        # "resultOffset": "",
        # "resultRecordCount": "",
        # "queryByDistance": "",
        # "returnExtentOnly": "false",
        # "datumTransformation": "",
        # "parameterValues": "",
        # "rangeValues": "",
        # "quantizationParameters": "",
        "featureEncoding": "esriDefault",
        "f": "geojson",
    }

    r = requests.get(url, params=payload)
    resp = r.json()

    # If the Rest Services has a 200 response
    if "features" in resp:
        # If the features are not empty, then the DEM exist
        if resp["features"] != []:
            return_val = True
        # If features are empty, then no DEM
        if resp["features"] == []:
            return_val = False
    # If 'error', then there was an unsuccessful request
    if "error" in resp:
        return_val = False
    return return_val


# The function to loop thru all resolutions and submit queries


def query_dems(
    shape_type: str, coords: List[Tuple[float, float]], width: int = 100
) -> Any:
    """Query 3DEPElevation.

     Index based on shape type and coordinates for available spatial resolution.

    Parameters
    ----------
    shape_type : str
        One of: point, line, polygon
    coords : List[Tuple[float, float]]
        List of coordinate tuples, for example: [(x,y), (x,y)].
    width : int, optional
        Width to buffer bounding box of shape, by default 100

    Returns
    -------
    Any
        Dictionary of query keys for example: res_1m, and bool intersection values.
    """
    resp = {}  # Create an empty dictionary for the response
    bbox = make_bbox(shape_type, coords, width)  # Make the bbox
    print(bbox)
    for (
        res_type
    ) in res_types:  # Loop thru all resolutions and submit a query for each one
        resp[res_type] = get_dem(bbox, res_type)  # Add the response to the dictionary

    # print(resp)
    return resp


def query_dems_bbox(bbox: Tuple[float, float, float, float]) -> Any:
    """Query 3DEP Elevation Index for available spatial resolution.

    Uses esriSpatialRelIntersects - Query Geometry Intersects Target Geometry.

    Parameters
    ----------
    bbox : Tuple[float]
        (minx, miny, maxx, maxy)

    Returns
    -------
    Dict
        Boolian values associated with resolution keys.
    """
    resp = {}

    for res_type in res_types:
        resp[res_type] = get_dem(bbox, res_type)
    return resp



"""Module containing XSGen class."""
from typing import Tuple

import geopandas as gpd
import numpy as np
import numpy.typing as npt
import pandas as pd
from shapely.geometry import LineString
from shapely.geometry import Point

from .centerline import Centerline


class XSGen:
    """The XSGen class.

    The XSGen class generates a cross-section line on an NHD stream segment.

    Attributes:
        point (GeoDataFrame): GeoDataFrame containing the point geometry of interest.
        cl_geom (GeoDataFrame): GeoDataFrame containing the NHD stream-segment geometry
            determined by following the rain-drop trace from the point to the NHD
            stream-segment.
        ny (int): Number of points in the cross-section.
        width (float): Width of the cross-section.
        tension (float): Tension used to generate a splined interpolation of the stream-
            segment geometry (>0 - 100.). Tension of 100 returns a linear interpolation.
    """

    def __init__(
        self: "XSGen",
        point: gpd.GeoDataFrame,
        cl_geom: gpd.GeoDataFrame,
        ny: int,
        width: float,
        tension: float = 10.0,
    ) -> None:
        """Constructor for XSGen class.

        Args:
            point (gpd.GeoDataFrame): GeoDataFrame containing the point geometry of
                interest.
            cl_geom (gpd.GeoDataFrame): GeoDataFrame containing the NHD stream-segment
                geometry determined by following the rain-drop trace from the point to
                the NHD stream-segment.
            ny (int): Number of points in the cross-section.
            width (float): Width of the cross-section.
            tension (float): Tension used to generate a splined interpolation
                of the stream-segment geometry (>0 - 100.). Tension of 100 returns a
                linear interpolation. Defaults to 10.0.
        """
        self.cl_geom = cl_geom
        self.point = point
        self.tension = tension
        self.width = width
        self.cl_length = self.cl_geom.geometry[0].length
        self.cl_npts = len(self.cl_geom.geometry[0].coords)
        if self.cl_length < 10.0:
            self.nx = 10
        else:
            self.nx = int(self.cl_length / 3)
        if self.nx % 2 == 0:
            self.nx += 1
        if ny % 2 == 0:
            ny += 1
        self.ny = ny
        # if self.cl_length > 20.0:
        #     self.nx = int(self.cl_length / 10)
        # else:
        #     self.nx = int(self.cl_length / 1)
        self.cl = Centerline(cl_geom, self.nx, self.tension)
        self.x = np.zeros(self.ny, dtype=np.double)
        self.y = np.zeros(self.ny, dtype=np.double)
        self._buildxs()

    def _get_perp_index(
        self: "XSGen", clx: npt.NDArray[np.double], cly: npt.NDArray[np.double]
    ) -> int:
        mind = 1e6
        id = -1
        for index, p in enumerate(zip(clx, cly)):
            dist = np.sqrt(
                np.power(self.point.geometry[0].x - p[0], 2)
                + np.power(self.point.geometry[0].y - p[1], 2)
            )
            if dist < mind:
                mind = dist
                id = index

        return id

    def _buildxs(self: "XSGen") -> None:
        clx, cly = self.cl.getinterppts()
        delt = np.double(self.width / (self.ny - 1))
        nm = int((self.ny + 1) / 2)
        index = self._get_perp_index(clx, cly)

        for id, j in enumerate(range(0, self.ny)):
            self.x[id] = clx[index] + delt * (nm - j - 1) * np.sin(
                self.cl.getphiinterp(index)
            )
            self.y[id] = cly[index] - delt * (nm - j - 1) * np.cos(
                self.cl.getphiinterp(index)
            )

    def get_xs(self: "XSGen") -> gpd.GeoDataFrame:
        """Get Geopandas DataFrame of generated cross-section.

        Returns:
            Geopandas DataFrame: Of generated cross-section
        """
        points = gpd.GeoSeries(map(Point, zip(self.x, self.y)))
        ls = LineString((points.to_list()))
        d = {0: {"name": "cross-section", "geometry": ls}}
        df = pd.DataFrame.from_dict(d, orient="index")
        gdf = gpd.GeoDataFrame(df, geometry=df.geometry, crs=self.point.crs)
        return gdf

    def get_xs_points(
        self: "XSGen",
    ) -> Tuple[npt.NDArray[np.double], npt.NDArray[np.double]]:
        """Get cross-section points.

        Returns:
            numpy array: Arrays of x and y points of cross-section.
        """
        return self.x, self.y

    def get_strm_seg_spline(self: "XSGen") -> gpd.GeoDataFrame:
        """Get the resulting splined stream-segment.

        Returns:
            Geopandas DataFrame: Of splined stream-segment.
        """
        x, y = self.cl.getinterppts()
        points = gpd.GeoSeries(map(Point, zip(x, y)))
        ls = LineString((points.to_list()))
        d = {0: {"name": "strm_seg_spline", "geometry": ls}}
        df = pd.DataFrame.from_dict(d, orient="index")
        gdf = gpd.GeoDataFrame(df, geometry=df.geometry)
        return gdf



"""pygeoapi implementation of getxsatendpts."""
import logging
import time

from geopandas import GeoDataFrame
from pygeoapi.process.base import BaseProcessor

from nldi_xstool.nldi_xstool import getxsatendpts

# from typing import Any
# from typing import Tuple


LOGGER = logging.getLogger(__name__)

PROCESS_METADATA = {
    "version": "0.1.0",
    "id": "nldi-xsatendpts",
    "title": "NLDI xsatendpts process",
    "description": "NLDI xsatendpts process",
    "keywords": ["NLDI xsatendpts"],
    "links": [
        {
            "type": "text/html",
            "rel": "canonical",
            "title": "information",
            "href": "https://example.org/process",
            "hreflang": "en-US",
        }
    ],
    "inputs": [
        {
            "id": "lat",
            "title": "lat",
            "input": {
                "literalDataDomain": {
                    "dataType": "list",
                    "valueDefinition": {"anyValue": True},
                }
            },
            "minOccurs": 2,
            "maxOccurs": 2,
        },
        {
            "id": "lon",
            "title": "lon",
            "input": {
                "literalDataDomain": {
                    "dataType": "list",
                    "valueDefinition": {"anyValue": True},
                }
            },
            "minOccurs": 2,
            "maxOccurs": 2,
        },
        {
            "id": "numpts",
            "title": "numpts",
            "input": {
                "literalDataDomain": {
                    "dataType": "int",
                    "valueDefinition": {"anyValue": True},
                }
            },
            "minOccurs": 1,
            "maxOccurs": 1,
        },
        {
            "id": "3dep_res",
            "title": "resolution",
            "abstract": "Resolution of 3dep elevation data",
            "minOccurs": 1,
            "maxOccurs": 1,
            "input": {
                "literalDataDomain": {
                    "dataType": "enum",
                    "valueDefinition": {
                        "anyValue": False,
                        "defaultValue": "10",
                        "possibleValues": ["30", "10", "5", "3", "1"],
                    },
                }
            },
        },
    ],
    "outputs": [
        {
            "id": "nldi-xsatendpts-response",
            "title": "output nldi-xsatendpts",
            "output": {"formats": [{"mimeType": "application/json"}]},
        }
    ],
    "example": {
        "inputs": [
            {"id": "lat", "value": [40.267720, 40.270568], "type": "text/plain"},
            {"id": "lon", "value": [-103.801086, -103.80097], "type": "text/plain"},
            {"id": "numpts", "value": "101", "type": "text/plain"},
            {"id": "3dep_res", "value": "1", "type": "text/plain"},
        ]
    },
}


class NLDIxsatendptsProcessor(BaseProcessor):  # type: ignore
    """NLDI xsatendpoints Processor."""

    def __init__(self, provider_def):  # type: ignore
        """Initialize object.

        :param provider_def: provider definition
        :returns: pygeoapi.process.nldi_delineate.NLDIDelineateProcessor
        """
        BaseProcessor.__init__(self, provider_def, PROCESS_METADATA)

    def execute(self, data):  # type: ignore
        """Execute processor."""
        print("before data assign")
        print(data)
        # reformat data into dict with just needed values
        newdata = {d["id"]: d["value"] for d in data}
        mimetype = "application/json"
        lat = newdata.get("lat")
        lon = newdata.get("lon")
        numpts = int(newdata.get("numpts"))  # type: ignore
        res = int(newdata.get("3dep_res"))  # type: ignore

        print(lat, lon, numpts, res)

        timebefore = time.perf_counter()

        print("before function")
        results = GeoDataFrame(
            getxsatendpts(
                path=[(lon[0], lat[0]), (lon[1], lat[1])],  # type: ignore
                numpts=numpts,
                res=res,
                crs="epsg:4326",
            )
        )
        print("after function")
        # print(results)

        timeafter = time.perf_counter()
        totaltime = timeafter - timebefore
        print("Total Time:", totaltime)

        # outputs = [
        #     {"id": "nldi-xsatendpts-response", "value": results.__geo_interface__}
        # ]
        # print(results)
        return mimetype, results.__geo_interface__

    def __repr__(self):  # type: ignore
        """Get representation."""
        return "<NLDIxsatendptsProcessor> {}".format("get xsatendpts")



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['nldi_xstool', 'nldi_xstool.process']

package_data = \
{'': ['*']}

install_requires = \
['Bottleneck>=1.3.2,<2.0.0',
 'OWSLib>=0.25.0,<0.26.0',
 'click<8',
 'dask>=2021.11.2,<2022.0.0',
 'dataretrieval>=0.7,<0.8',
 'lxml>=4.6.4,<5.0.0',
 'netCDF4>=1.5.8,<2.0.0',
 'numpy>=1.21.0,<2.0.0',
 'py3dep',
 'pygeoapi>=0.11.0,<0.12.0',
 'pygeohydro',
 'pygeoogc==0.11.4',
 'pygeoutils',
 'pynhd',
 'scipy>=1.7.3,<2.0.0',
 'xarray>=0.20.1,<0.21.0']

entry_points = \
{'console_scripts': ['nldi-xstool = nldi_xstool.__main__:main']}

setup_kwargs = {
    'name': 'nldi-xstool',
    'version': '0.12.0.dev0',
    'description': 'Nldi Xstool',
    'long_description': "# Nldi Xstool\n\n[![PyPI](https://img.shields.io/pypi/v/nldi-xstool.svg)](https://pypi.org/project/nldi-xstool/)\n[![Status](https://img.shields.io/pypi/status/nldi-xstool.svg)](https://pypi.org/project/nldi-xstool/)\n[![Python Version](https://img.shields.io/pypi/pyversions/nldi-xstool)](https://pypi.org/project/nldi-xstool)\n[![License](https://img.shields.io/pypi/l/nldi-xstool)](https://creativecommons.org/publicdomain/zero/1.0/legalcode)\n\n[![Read the documentation at https://nldi-xstool.readthedocs.io/](https://img.shields.io/readthedocs/nldi-xstool/latest.svg?label=Read%20the%20Docs)](https://nldi-xstool.readthedocs.io/)\n\n[![pipeline status](https://code.usgs.gov/wma/nhgf/toolsteam/nldi-xstool/badges/main/pipeline.svg)](https://code.usgs.gov/wma/nhgf/toolsteam/nldi-xstool/-/commits/main)\n[![coverage report](https://code.usgs.gov/wma/nhgf/toolsteam/nldi-xstool/badges/main/coverage.svg)](https://code.usgs.gov/wma/nhgf/toolsteam/nldi-xstool/-/commits/main)\n\n[![pre-commit](https://img.shields.io/badge/pre--commit-enabled-brightgreen?logo=pre-commit&logoColor=white)](https://github.com/pre-commit/pre-commit)\n[![Black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n[![Poetry](https://img.shields.io/badge/poetry-enabled-blue)](https://python-poetry.org/)\n[![Conda](https://img.shields.io/badge/conda-enabled-green)](https://anaconda.org/)\n\n## Features\n\n- TODO\n\n## Requirements\n\n- TODO\n\n## Installation\n\nYou can install nldi-xstool via\n[pip](https://pip.pypa.io/) from [PyPI](https://pypi.org/):\n\n```python\npip install nldi-xstool\n```\n\n## Command Line Usage\n\nPlease see the {doc}`usage`\n\n## Contributing\n\nContributions are very welcome.\nTo learn more, see the {doc}`contributing`.\n\n## License\n\nDistributed under the terms of the [CC0 1.0 Universal license](https://creativecommons.org/publicdomain/zero/1.0/legalcode),\n_Nldi Xstool_ is free and open source software.\n\n## Disclaimer\n\n{doc}`disclaimer`\n\n## Issues\n\nIf you encounter any problems,\nplease [file an issue](https://code.usgs.gov/wma/nhgf/toolsteam/nldi-xstool/-/issues) along with a detailed description.\n\n## Credits\n\nThis project was generated from\n[@hillc-usgs](https://github.com/hillc-usgs)'s [Pygeoapi Plugin\nCookiecutter](https://code.usgs.gov/wma/nhgf/pygeoapi-plugin-cookiecutter)\ntemplate.\n",
    'author': 'Richard McDonald',
    'author_email': 'rmcd@usgs.gov',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://code.usgs.gov/wma/nhgf/toolsteam/nldi-xstool',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9.0,<3.10.0',
}


setup(**setup_kwargs)
